/*
 * FishFight render engine
 * Licensed under the BSD license, see LICENSE in source root for details.
 */

#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspctrl.h>
#include <pspdebug.h>
#include <psprtc.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#include <deque>

#include <pspgu.h>
#include <pspgum.h>

#include <mikmod.h>

#include "callbacks.h"
#include "vram.h"
#include "graphics.h"
#include "mesh.h"
#include "connect4.h"
#include "mjpegplayer.h"

PSP_MODULE_INFO("PSP Connect 4", 0, 1, 1);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER);

static unsigned int __attribute__((aligned(16))) gulist[262144];

int irfd=-1;	// Infra Red file descriptor

long int timeGetTime()
{
	u64 tick;
	sceRtcGetCurrentTick(&tick);

	return (long)tick/1000;
}

void swizzleFast(Image *source)
{
   unsigned int width=source->textureWidth*4;
   unsigned int height=source->textureHeight;
   u32* out=(u32 *)memalign(16,width*height);
   unsigned int blockx, blocky;
   int i;
 
   unsigned int widthBlocks = (width / 16);
   unsigned int heightBlocks = (height / 8);
 
   unsigned int srcPitch = (width-16)/4;
   unsigned int srcRow = width * 8;
 
   const u8* ysrc = (u8 *)source->data;;
   u32* dst = out;
 
   for (blocky = 0; blocky < heightBlocks; ++blocky) {
      const u8* xsrc = ysrc;
      for (blockx = 0; blockx < widthBlocks; ++blockx) {
         const u32* src = (u32*)xsrc;
         for (i=0;i<8;i++) {
            *(dst++) = *(src++);
            *(dst++) = *(src++);
            *(dst++) = *(src++);
            *(dst++) = *(src++);
            src += srcPitch;
         }
         xsrc += 16;
     }
     ysrc += srcRow;
   }
   free(source->data);
   source->data=(Color *)out;
   source->swizzled=1;
}

void drawMesh(Mesh *mesh)
{
#if FF_TEXTURE_ENABLE
	if(mesh->mat.size()>0 && mesh->mat.begin()->textureName.size()>0) {
		Image *image=mesh->mat.begin()->texture;
		if(!image) {
			printf("Mesh missing texture %s\n",mesh->mat.begin()->textureName.c_str());
		} else {
			if(image->swizzled==0) {
				swizzleFast(image);
			}
			sceGuTexMode(GU_PSM_8888,0,0,image->swizzled);
			sceGuTexImage(0,image->imageWidth,image->imageHeight,image->textureWidth,image->data);
		}
	}
	sceGumDrawArray(GU_TRIANGLES,GU_TEXTURE_32BITF|GU_NORMAL_32BITF|GU_VERTEX_32BITF|GU_INDEX_16BIT|GU_TRANSFORM_3D,mesh->polyCount*3,mesh->poly,mesh->vert);
#else
#if FF_VERTEXCOLOR_ENABLE
	sceGumDrawArray(GU_TRIANGLES,GU_NORMAL_32BITF|GU_COLOR_8888|GU_VERTEX_32BITF|GU_INDEX_16BIT|GU_TRANSFORM_3D,mesh->polyCount*3,mesh->poly,mesh->vert);
#else
	sceGuColor(0xff4488);
	sceGumDrawArray(GU_TRIANGLES,GU_NORMAL_32BITF|GU_VERTEX_32BITF|GU_INDEX_16BIT|GU_TRANSFORM_3D,mesh->polyCount*3,mesh->poly,mesh->vert);
#endif
#endif
}

void drawFrame(Frame *frame)
{
	std::vector<Frame *>::iterator q;
	std::vector<Mesh *>::iterator p;

	sceGumMatrixMode(GU_MODEL);
	sceGumPushMatrix();

	sceGumMultMatrix((ScePspFMatrix4 *)frame->transform);

	for(q=frame->frame.begin();q!=frame->frame.end();q++) {
		Frame *frame=*q;
		drawFrame(frame);
	}
	for(p=frame->mesh.begin();p!=frame->mesh.end();p++) {
		Mesh *mesh=*p;
		drawMesh(mesh);
	}
	sceGumMatrixMode(GU_MODEL);
	sceGumPopMatrix();
}

void drawEntity(Entity *entity)
{
	std::vector<Frame *>::iterator p;

	if(!entity) return;

	for(p=entity->frame.begin();p!=entity->frame.end();p++) {
		Frame *frame=*p;
		drawFrame(frame);
	}
}

void blitAlphaImageToScreen(int sx, int sy, int width, int height, Image* source, int dx, int dy)
{
	if(source->swizzled==0) {
		swizzleFast(source);
	}
	sceGuTexMode(GU_PSM_8888,0,0,source->swizzled);
	sceGuTexImage(0, source->textureWidth, source->textureHeight, source->textureWidth, (void*) source->data);
	float u = 1.0f / ((float)source->textureWidth);
	float v = 1.0f / ((float)source->textureHeight);
	sceGuTexScale(u, v);
	sceGuDisable(GU_DEPTH_TEST);
	
	int j = 0;
	while (j < width) {
		struct Vertex {
			unsigned short u,v;
			short x,y,z;
		}* vertices = (Vertex*) sceGuGetMemory(2 * sizeof(Vertex));
		int sliceWidth = 64;
		if (j + sliceWidth > width) sliceWidth = width - j;
		vertices[0].u = sx + j;
		vertices[0].v = sy;
		vertices[0].x = dx + j;
		vertices[0].y = dy;
		vertices[0].z = 0;
		vertices[1].u = sx + j + sliceWidth;
		vertices[1].v = sy + height;
		vertices[1].x = dx + j + sliceWidth;
		vertices[1].y = dy + height;
		vertices[1].z = 0;
		sceGuDrawArray(GU_SPRITES, GU_TEXTURE_16BIT | GU_VERTEX_16BIT | GU_TRANSFORM_2D, 2, 0, vertices);
		j += sliceWidth;
	}

	sceGuEnable(GU_DEPTH_TEST);
	sceGuTexScale(1.0, 1.0);
	sceGuTexMode(GU_PSM_8888,0,0,0);
}

Connect4 fishFight;
void *drawBuffer;
#define LIGHT_DISTANCE 3
float camerax=40,cameray=5,cameraz=-10;

// Mesages are: 0=menu, 1=player1win, 2=player2win 3=tie
enum { M_MENU, M_WIN1, M_WIN2, M_TIE, M_CURSOR, M_HUDRED, M_HUDYELLOW, M_WAITING, M_PAUSESOUND, M_PAUSENOSOUND, M_CPUWIN, MAX_MESSAGE };
Image *message[MAX_MESSAGE];
char *messagepath[MAX_MESSAGE]={ "menu.png","win1.png","win2.png","tie.png","cursor.png","hudred.png","hudyellow.png","waiting.png","pause1.png","pause2.png","cpuwin.png"};
unsigned long currentFrame,oldFrame;
 
void display()
{
	currentFrame=timeGetTime();
	sceGuStart(GU_DIRECT,gulist);

	// clear screen
	sceGuClearColor(0xff554433);
	sceGuClearDepth(0);
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	sceGumPerspective(45.0f,16.0f/9.0f,1.5f,512.0f);

	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();

#if 0	
	{
		ScePspFVector3 pos = { 0, 0, -2.5f };
		//ScePspFVector3 rot = { val * 0.79f * (GU_PI/180.0f), val * 0.98f * (GU_PI/180.0f), val * 1.32f * (GU_PI/180.0f) };
		ScePspFVector3 rot= { -90*GU_PI/180.0f,0,0};
		ScePspFVector3 scale = { 0.25f, 0.25f, 0.25f };
		pos.z*=9;
		sceGumTranslate(&pos);
		sceGumRotateXYZ(&rot);
		sceGumScale(&scale);
	}
#else
	ScePspFVector3 from={camerax,cameray,cameraz};
	
	ScePspFVector3 to= { 0,-5,-10 };
	ScePspFVector3 up= { 0,1,0 };
	sceGumLookAt(&from,&to,&up);
#endif
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();

#if 1
	std::vector<Piece *>::iterator p;
	for(p=fishFight.piece.begin();p!=fishFight.piece.end();p++) {
		Piece *piece=*p;
		sceGumPushMatrix();
		sceGumMultMatrix((ScePspFMatrix4 *)piece->getTransform());
		drawEntity(piece->getModel());
		sceGumPopMatrix();
	}
#endif

	fishFight.updateLevel(currentFrame);

	Image *overlay=0;
	// Now overlay messages
	int msgId=-1;
	if(fishFight.activePlayer==BP_RED && fishFight.mode!=GM_CLIENT) msgId=M_HUDRED;
	if(fishFight.activePlayer==BP_YELLOW && fishFight.mode!=GM_PVSAI && fishFight.mode!=GM_SERVER) msgId=M_HUDYELLOW;
	if(fishFight.mode==GM_MODEMENU) msgId=M_MENU;
	if(fishFight.mode==GM_REDWINS) msgId=M_WIN1;
	if(fishFight.mode==GM_YELLOWWINS) msgId=M_WIN2;
	if(fishFight.mode==GM_TIE) msgId=M_TIE;
	if(fishFight.mode==GM_SERVER && fishFight.activePlayer==BP_YELLOW) msgId=M_WAITING;
	if(fishFight.mode==GM_CLIENT && fishFight.activePlayer==BP_RED) msgId=M_WAITING;
	if(fishFight.mode==GM_CPUWINS) msgId=M_CPUWIN;
	if(fishFight.pauseMenu) msgId=fishFight.sound?M_PAUSESOUND:M_PAUSENOSOUND;
	if(msgId>=0){
		if(!message[msgId]) message[msgId]=loadImage(messagepath[msgId]);
		overlay=message[msgId];
		if(overlay) {
			blitAlphaImageToScreen(0, 0, overlay->imageWidth, overlay->imageHeight, overlay, 0, 0);
			//sceGuCopyImage(GU_PSM_8888, 0, 0, overlay->imageWidth, overlay->imageHeight, overlay->textureWidth, overlay->data, 0, 0, PSP_LINE_SIZE, (Color *)((unsigned int)drawBuffer+0x4000000));
		}
		if(fishFight.mode==GM_MODEMENU || fishFight.pauseMenu) {
			// Draw the option pointer.
			if(!message[4]) message[4]=loadImage(messagepath[4]);
			overlay=message[4];
			if(overlay && fishFight.pauseMenu) {
				blitAlphaImageToScreen(0,0,overlay->imageWidth, overlay->imageHeight, overlay, 95, fishFight.getArrowColumn()*30+90);
			} else if(overlay) {
				blitAlphaImageToScreen(0,0,overlay->imageWidth, overlay->imageHeight, overlay, 25, fishFight.getArrowColumn()*45+30);
			}
		}
	}

	sceGuFinish();
	sceKernelDcacheWritebackAll();
	sceGuSync(0,0);

	pspDebugScreenSetXY(0,0);
	//Piece *piece=*(fishFight.piece.begin());
	//float *t=piece->getTransform();
	//pspDebugScreenPrintf("piece: %.2f,%.2f,%.2f\n",t[12],t[13],t[14]);
	//pspDebugScreenPrintf("Frame count %d\n",piece->getModel()->frame.size());
//	pspDebugScreenPrintf("Camera: %.3f %.3f %.3f\n",camerax,cameray,cameraz);
#if 0
	if(piece->getModel()->frame.size()>0) {
		Frame *f=*piece->getModel()->frame.begin();
		pspDebugScreenPrintf("Mesh count %d\n",f->mesh.size());
		if(f->mesh.size()>0) {
			Mesh *m=*f->mesh.begin();
			pspDebugScreenPrintf("Mesh verts: %d, polys: %d\n",m->vertCount,m->polyCount);
		}
	}
#endif
	int elapsed=currentFrame-oldFrame;
	if(elapsed>1000) elapsed=1000;
	if(elapsed<1) elapsed=1;
//	pspDebugScreenPrintf("FPS %d\n",1000/(oldFrame-currentFrame));
	oldFrame=currentFrame;

	sceDisplayWaitVblankStart();	// Probably wastes some cycles here.
	drawBuffer=sceGuSwapBuffers();
	pspDebugScreenSetOffset((int)drawBuffer);
}	

#define BUF_WIDTH (512)
#define SCR_WIDTH (480)
#define SCR_HEIGHT (272)

#define MAXSONG 5

SAMPLE *sfx[S_MAX];
UNIMOD *song[MAXSONG];
char *sfxpath[]={
	"intro.wav",
	"start.wav",
	"winner.wav",
	"computerwon.wav",
	"droppiece.wav",
	"badmove.wav",
	"arrow.wav",
};

void mikmodErrorHandler(void)
{
	printf("_mm_critical %d\n", _mm_critical);
	printf("_mm_errno %d\n", _mm_errno);
	printf("%s\n", _mm_errmsg[_mm_errno]);
	return;
}

void playSound(int id)
{
	if(fishFight.sound==false) return;
	if(id>=0 && id<S_MAX && sfx[id]!=0) {
		int voice = MikMod_PlaySample(sfx[id],0,0);
		printf("playing %s on voice %d\n",sfxpath[id],voice);
	}
	return;
}

std::deque<unsigned char> broadcastQueue;
time_t lastResent=0;

void broadcast(unsigned char update)
{
	std::deque<unsigned char>::iterator p;

	if(irfd>0) {
		for(p=broadcastQueue.begin();p!=broadcastQueue.end();p++) {
			unsigned char ch=*p;
			printf("Resending message 0x02x\n",ch);
			sceIoWrite(irfd,&ch,1);
		}
		printf("Sending message 0x02x\n",update);
		sceIoWrite(irfd,&update,1);
		broadcastQueue.push_back(update);
	}
}

unsigned int colors[4] = { 0xffff0000, 0xff00ff00, 0xff0000ff, 0xffff00ff };

int main(int argc, char* argv[])
{
	SceCtrlData pad;

	setupCallbacks();

	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(1);

	_mm_RegisterErrorHandler(mikmodErrorHandler);
	MikMod_RegisterAllLoaders();
	MikMod_RegisterAllDrivers();
	md_mode = DMODE_16BITS|DMODE_STEREO|DMODE_SOFT_SNDFX|DMODE_SOFT_MUSIC;
	md_reverb = 0;
	md_pansep = 128;
	MikMod_Init();
	MikMod_SetNumVoices(-1, 8);

	Player_Start(NULL);

	int i;
	for(i=0;i<S_MAX;i++) {
		printf("reading sound file %d...\n",i);
		FILE *file=fopen(sfxpath[i],"rb");
		if(file) {
			printf("opening file %s\n",sfxpath[i]);
			fclose(file);
			sfx[i]=WAV_LoadFN(sfxpath[i]);
		}
		if(!sfx[i]) printf("Warning: missing sound file '%s'\n",sfxpath[i]);
	}

	playSound(S_INTRO);
	playMJPEG("intro.mjpeg",30);

	// setup GU

	void* fbp0 = getStaticVramBuffer(BUF_WIDTH,SCR_HEIGHT,GU_PSM_8888);
	void* fbp1 = getStaticVramBuffer(BUF_WIDTH,SCR_HEIGHT,GU_PSM_8888);
	void* zbp = getStaticVramBuffer(BUF_WIDTH,SCR_HEIGHT,GU_PSM_4444);

	Image *source=loadImage("controls.png");

	sceGuInit();
	sceGuStart(GU_DIRECT,gulist);
	sceGuDrawBuffer(GU_PSM_8888,fbp0,BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH,SCR_HEIGHT,fbp1,BUF_WIDTH);
	sceGuClearColor(0xff774433);
	sceGuClear(GU_COLOR_BUFFER_BIT);
#if 1
	if(source) 
		sceGuCopyImage(GU_PSM_8888, 0, 0, source->imageWidth, source->imageHeight, source->textureWidth, source->data, 0, 0, PSP_LINE_SIZE, (Color *)((unsigned int)drawBuffer+0x4000000));
#else
	if(source) 
		blitAlphaImageToScreen(0, 0, source->imageWidth, source->imageHeight, source, 0, 0);
#endif
	sceGuFinish();
	sceGuSync(0,0);

	sceDisplayWaitVblankStart();
	drawBuffer=sceGuSwapBuffers();
	sceGuDisplay(GU_TRUE);

	//pspDebugScreenInit();
	//pspDebugScreenPrintf("PSP Connect 4\nLoading models...\n");
	fishFight.newGame();
	fishFight.load(1);

	if(source) freeImage(source);

#if 0
	source=loadImage("51063_splashpsp.png");
	if(source) {
		sceGuStart(GU_DIRECT,gulist);
		sceGuCopyImage(GU_PSM_8888, 0, 0, source->imageWidth, source->imageHeight, source->textureWidth, source->data, 0, 0, PSP_LINE_SIZE, (Color *)((unsigned int)drawBuffer+0x4000000));
		
		sceGuFinish();
		sceGuSync(0,0);
		sceDisplayWaitVblankStart();
		drawBuffer=sceGuSwapBuffers();

		// Now wait a while
		int i;
		for(i=0;i<120;i++) {
			sceDisplayWaitVblankStart();
		}
	}
	if(source) freeImage(source);
#endif
	source=0;
#if 0
	pspDebugScreenPrintf("Press [X] to start.\n");

	do {
		sceDisplayWaitVblankStart();
		sceCtrlReadBufferPositive(&pad,1);
	} while(!(pad.Buttons & PSP_CTRL_CROSS));
	// run sample

	do {
		sceDisplayWaitVblankStart();
		sceCtrlReadBufferPositive(&pad,1);
	} while((pad.Buttons & PSP_CTRL_CROSS));
#endif

#if 1

	sceGuStart(GU_DIRECT,gulist);
	sceGuDrawBuffer(GU_PSM_8888,fbp0,BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH,SCR_HEIGHT,fbp1,BUF_WIDTH);
	sceGuDepthBuffer(zbp,BUF_WIDTH);
	sceGuOffset(2048 - (SCR_WIDTH/2),2048 - (SCR_HEIGHT/2));
	sceGuViewport(2048,2048,SCR_WIDTH,SCR_HEIGHT);
	sceGuDepthRange(65535,0);
	sceGuScissor(0,0,SCR_WIDTH,SCR_HEIGHT);
	sceGuEnable(GU_SCISSOR_TEST);
	sceGuDepthFunc(GU_GEQUAL);
	sceGuEnable(GU_DEPTH_TEST);
	sceGuFrontFace(GU_CCW);
	sceGuShadeModel(GU_SMOOTH);
	sceGuEnable(GU_CULL_FACE);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuEnable(GU_CLIP_PLANES);
	//sceGuDisable(GU_CLIP_PLANES);
#ifdef FF_TEXTURE_ENABLE
	sceGuTexMode(GU_PSM_8888,0,0,0);
	sceGuEnable(GU_BLEND);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	//sceGuTexFunc(GU_TFX_ADD,GU_TCC_RGB);
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	sceGuTexEnvColor(0xffff00);
	sceGuTexFilter(GU_LINEAR,GU_LINEAR);
	sceGuTexScale(1.0f,1.0f);
#endif
#if 1
        sceGuEnable(GU_LIGHTING);
	sceGuEnable(GU_LIGHT0);
	sceGuEnable(GU_LIGHT1);
	sceGuEnable(GU_LIGHT2);
	sceGuEnable(GU_LIGHT3);
#endif
	sceGuFinish();
	sceGuSync(0,0);

	sceDisplayWaitVblankStart();
	sceGuDisplay(GU_TRUE);

	int oldButtons=0;

	while(running())
	{
		int changedButtons;
		display();

		if(fishFight.mode==GM_SERVER ||fishFight.mode==GM_CLIENT) {
			if(irfd<=0) {
				irfd=sceIoOpen("irda0:", PSP_O_RDWR,0);
			}
			if(irfd>0 && broadcastQueue.size()>0 && lastResent++>60) {
				std::deque<unsigned char>::iterator p;
				for(p=broadcastQueue.begin();p!=broadcastQueue.end();p++) {
					unsigned char ch=*p;
					printf("Timeout.  Resending message 0x02x\n",ch);
					sceIoWrite(irfd,&ch,1);
				}
				lastResent=0;
			}

			// Poll for the incoming message
			unsigned char buf[1];
			buf[0]=0x7f;
			int len=sceIoRead(irfd,buf,1);
			// So, trigger event
			if(len>0 && (buf[0]&0x40)==0) {
				int down=buf[0]&0x80?1:0;
				fishFight.buttonEvent(buf[0]&0x7f,down,(int)time);
				printf("Received message 0x02x\n",buf[0]);
				buf[0]|=0x40;	// Send Acknowledgement.
				sceIoWrite(irfd,buf,1);
				printf("Sending ack message 0x02x\n",buf[0]);
			} else if(len>0) {
				// Process acknowledgement.
				int msg=buf[0]&0xbf;
				printf("Received ack message 0x02x\n",buf[0]);
				
				std::deque<unsigned char>::iterator p;
				for(p=broadcastQueue.begin();p!=broadcastQueue.end();p++) {
					int ch=*p;
					if(ch==msg) {
						broadcastQueue.erase(p);
						printf("Matched ack message 0x02x\n",buf[0]);
						break;
					}
				}	// Or maybe couldn't find.
			}
		} else {
			if(irfd>0) sceIoClose(irfd);
			irfd=-1;
		}
		sceCtrlReadBufferPositive(&pad,1);
		changedButtons=oldButtons ^ pad.Buttons;
		unsigned long time=timeGetTime();
		static float angle=0.1f;
		if(pad.Buttons & PSP_CTRL_LTRIGGER) {
			angle+=0.1f;
			if(angle>M_PI/2.0f) angle=M_PI/2.0f;
			camerax=cosf(angle)*40;
			cameraz=sinf(angle)*40-10;
		}
		if(pad.Buttons & PSP_CTRL_RTRIGGER) {
			angle-=0.1f;
			if(angle<-M_PI/2.0f) angle=-M_PI/2.0f;
			camerax=cosf(angle)*40;
			cameraz=sinf(angle)*40-10;
		}
#if 1	
		for (i = 0; i < 4; ++i) {
			ScePspFVector3 pos = { cosf(i*(GU_PI/2) - angle ) * LIGHT_DISTANCE, 10, sinf(i*(GU_PI/2) - angle ) * LIGHT_DISTANCE
			};
			sceGuLight(i,GU_POINTLIGHT,GU_DIFFUSE_AND_SPECULAR,&pos);
			sceGuLightColor(i,GU_DIFFUSE,colors[i]);
			sceGuLightColor(i,GU_SPECULAR,0xffffffff);
			sceGuLightAtt(i,0.0f,1.0f,0.0f);

		}
		sceGuSpecular(12.0f);
		sceGuAmbient(0x00222222);
#endif
		if(fishFight.mode!=GM_CLIENT) {
			if(changedButtons & PSP_CTRL_DOWN) {
				fishFight.buttonEvent(FF_DOWN,(pad.Buttons&PSP_CTRL_DOWN)!=0,time);
				if(fishFight.mode==GM_SERVER && irfd>0) broadcast( FF_DOWN|((pad.Buttons&PSP_CTRL_DOWN)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_UP) {
				fishFight.buttonEvent(FF_UP,(pad.Buttons&PSP_CTRL_UP)!=0,time);
				if(fishFight.mode==GM_SERVER && irfd>0) broadcast( FF_UP|((pad.Buttons&PSP_CTRL_UP)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_LEFT) {
				fishFight.buttonEvent(FF_LEFT,(pad.Buttons&PSP_CTRL_LEFT)!=0,time);
				if(fishFight.mode==GM_SERVER && irfd>0) broadcast( FF_LEFT|((pad.Buttons&PSP_CTRL_LEFT)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_RIGHT) {
				fishFight.buttonEvent(FF_RIGHT,(pad.Buttons&PSP_CTRL_RIGHT)!=0,time);
				if(fishFight.mode==GM_SERVER && irfd>0) broadcast( FF_RIGHT|((pad.Buttons&PSP_CTRL_RIGHT)?0x80:0));
			}
		}
		if(fishFight.mode!=GM_SERVER) {
			if(changedButtons & PSP_CTRL_SQUARE) {
				fishFight.buttonEvent(FF_SQUARE,(pad.Buttons&PSP_CTRL_SQUARE)!=0,time);
				if(fishFight.mode==GM_CLIENT && irfd>0) broadcast( FF_SQUARE|((pad.Buttons&PSP_CTRL_SQUARE)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_CIRCLE) {
				fishFight.buttonEvent(FF_CIRCLE,(pad.Buttons&PSP_CTRL_CIRCLE)!=0,time);
				if(fishFight.mode==GM_CLIENT && irfd>0) broadcast( FF_CIRCLE|((pad.Buttons&PSP_CTRL_CIRCLE)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_TRIANGLE) {
				fishFight.buttonEvent(FF_TRIANGLE,(pad.Buttons&PSP_CTRL_TRIANGLE)!=0,time);
				if(fishFight.mode==GM_CLIENT && irfd>0) broadcast( FF_TRIANGLE|((pad.Buttons&PSP_CTRL_TRIANGLE)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_CROSS) {
				fishFight.buttonEvent(FF_CROSS,(pad.Buttons&PSP_CTRL_CROSS)!=0,time);
				if(fishFight.mode==GM_CLIENT && irfd>0) broadcast( FF_CROSS|((pad.Buttons&PSP_CTRL_CROSS)?0x80:0));
			}
		}
		if(fishFight.pauseMenu && fishFight.mode==GM_SERVER) {
			if(changedButtons & PSP_CTRL_CROSS) {
				fishFight.buttonEvent(FF_CROSS,(pad.Buttons&PSP_CTRL_CROSS)!=0,time);
				if(fishFight.mode==GM_CLIENT && irfd>0) broadcast( FF_CROSS|((pad.Buttons&PSP_CTRL_CROSS)?0x80:0));
			}
		}
		if(fishFight.pauseMenu && fishFight.mode==GM_CLIENT) {
			if(changedButtons & PSP_CTRL_DOWN) {
				fishFight.buttonEvent(FF_DOWN,(pad.Buttons&PSP_CTRL_DOWN)!=0,time);
				if(fishFight.mode==GM_SERVER && irfd>0) broadcast( FF_DOWN|((pad.Buttons&PSP_CTRL_DOWN)?0x80:0));
			}
			if(changedButtons & PSP_CTRL_UP) {
				fishFight.buttonEvent(FF_UP,(pad.Buttons&PSP_CTRL_UP)!=0,time);
				if(fishFight.mode==GM_SERVER && irfd>0) broadcast( FF_UP|((pad.Buttons&PSP_CTRL_UP)?0x80:0));
			}
		}
		if(changedButtons & PSP_CTRL_SELECT) {
			fishFight.buttonEvent(FF_SELECT,(pad.Buttons&PSP_CTRL_SELECT)!=0,time);
			if((pad.Buttons&PSP_CTRL_SELECT)==0) {
				// Screen shot.
				static int shotNo=0;
				char fname[256];
				shotNo++;
				sprintf(fname,"pc4-%02d.png",shotNo);
				saveImage(fname,(Color *)(0x44000000+(char *)drawBuffer),480,272,512,0);
			}
		}
		if(changedButtons & PSP_CTRL_START) {
			fishFight.buttonEvent(FF_START,(pad.Buttons&PSP_CTRL_START)!=0,time);
		}
		if(changedButtons & PSP_CTRL_HOLD) {
			fishFight.buttonEvent(FF_HOLD,(pad.Buttons&PSP_CTRL_HOLD)!=0,time);
		}
		oldButtons=pad.Buttons;	
		if(fishFight.mode==GM_QUIT) break;
	}

	sceGuTerm();
#endif

	sceKernelExitGame();
	return 0;
}
